﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using Dapfor.Wpf.Controls;
using Dapfor.Wpf.Grid.Benchmark.Utils;
using Brushes = Dapfor.Wpf.Grid.Benchmark.Utils.Brushes;

namespace Dapfor.Wpf.Grid.Benchmark
{
    /// <summary>
    /// Interaction logic for DataUpdatingEventDrivenModelControl.xaml
    /// </summary>
    public partial class DataUpdatingEventDrivenModelControl : UserControl, IActivable
    {
        private readonly PerformanceCounter _cpuCounter = new PerformanceCounter();
        private readonly Random _random = new Random();
        private readonly List<int> _updateScenarios = new List<int>();
        private DateTime _lastStat = DateTime.UtcNow;
        private int _lastStatCount;
        private DateTime _measureStart;
        private bool _measuring;
        private int _updatesCount;
        readonly List<TestPublisherDataObject> _dataObjects = new List<TestPublisherDataObject>();


        public DataUpdatingEventDrivenModelControl()
        {
            InitializeComponent();

            //Initialize performance counters
            _cpuCounter.CategoryName = "Processor";
            _cpuCounter.CounterName = "% Processor Time";
            _cpuCounter.InstanceName = "_Total";

            grid.ProcessUpdate += OnUpdateCell;

            grid.SettingsBlinking.Enabled = false;
            grid.RowUpdated += delegate(object sender, GridControlRowUpdateEventArgs e)
            {
                Brush color = _random.Next()%2 == 0 ? Brushes.Green : Brushes.Red;
                e.Row[e.DataField.Id].Blink(TimeSpan.FromMilliseconds(1000), color);

            };
        }

        #region IActivable Members

        public void Activate()
        {
            panel.IsEnabled = true;
        }

        public void Deactivate()
        {
            panel.IsEnabled = false;
            grid.StopLoadBalancingProcess();
            _updatesCount = 0;
            _measuring = true;
            _measureStart = DateTime.UtcNow;
            _lastStatCount = 0;
            _lastStat = DateTime.UtcNow;
            gbStatistics.Visibility = Visibility.Hidden;
            grid.Rows.Clear();
        }

        #endregion

        private void OnUpdateCell(object sender, EventArgs e)
        {
            if (grid.IsLimitFound && !_measuring)
            {
                _updatesCount = 0;
                _measuring = true;
                _measureStart = DateTime.UtcNow;
                _lastStatCount = 0;
                _lastStat = DateTime.UtcNow;
            }


            //Updates random cell in the grid;
            TestPublisherDataObject dataObject = _dataObjects.Count > 0 ? _dataObjects[_random.Next() % _dataObjects.Count] : null;
            int field = _random.Next() % grid.Headers[0].Count;

            if (dataObject != null)
            {
                int value = _random.Next();
                if (cbNonGuiThread.IsChecked == true)
                {
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        UpdateDataobject(dataObject, field, value);
                    });
                }
                else
                {
                    UpdateDataobject(dataObject, field, value);
                }
            }

            _updatesCount++;
            if (_measuring)
            {
                //Measure refresh rate for 5 seconds
                if (DateTime.UtcNow - _measureStart > TimeSpan.FromSeconds(5))
                {
                    grid.StopLoadBalancingProcess();

                    int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                                          ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                                          : 0;
                    long visibleCells = grid.Headers[0].Count*visibleRows;
                    long totalCells = grid.Headers[0].Count*grid.Rows.Count;

                    double rate = totalCells != 0 ? (double) visibleCells/totalCells : 0;


                    Logger.Log(string.Format("Grid of {0} rows, size = {2:0}x{3:0}. Maximum refresh rate: {1} updates/sec, in visible bounds: {4:0} updates/sec",
                                             grid.Rows.Count,
                                             _updatesCount/5,
                                             grid.ActualWidth,
                                             grid.ActualHeight,
                                             rate*_updatesCount/5));

                    //Execute the next scenario
                    StartTest();
                }
            }

            DisplayStatistics();
        }

        private static void UpdateDataobject(TestPublisherDataObject dataObject, int field, int value)
        {
            if (dataObject != null)
            {
                switch (field)
                {
                    case 0:
                        dataObject.Value0 = value;
                        break;
                    case 1:
                        dataObject.Value1 = value;
                        break;
                    case 2:
                        dataObject.Value2 = value;
                        break;
                    case 3:
                        dataObject.Value3 = value;
                        break;
                    case 4:
                        dataObject.Value4 = value;
                        break;
                    case 5:
                        dataObject.Value5 = value;
                        break;
                    case 6:
                        dataObject.Value6 = value;
                        break;
                    case 7:
                        dataObject.Value7 = value;
                        break;
                    case 8:
                        dataObject.Value8 = value;
                        break;
                    case 9:
                        dataObject.Value9 = value;
                        break;
                }
            }
        }

        private void DisplayStatistics()
        {
            if (DateTime.UtcNow - _lastStat > TimeSpan.FromMilliseconds(500))
            {
                lbCPULoadRate.Text = string.Format("CPU load rate: {0:0} %", _cpuCounter.NextValue());
                lbGridSize.Text = string.Format("Grid size: {0:0}x{1:0}", grid.ActualWidth, grid.ActualHeight);
                double updates = 1000*(_updatesCount - _lastStatCount)/(DateTime.UtcNow - _lastStat).TotalMilliseconds;
                lbUpdateRate.Text = string.Format("Update rate: {0:0} updates/sec", updates);

                int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                                      ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                                      : 0;
                long visibleCells = grid.Headers[0].Count*visibleRows;
                long totalCells = grid.Headers[0].Count*grid.Rows.Count;

                double rate = totalCells != 0 ? (double) visibleCells/totalCells : 0;
                lbVisibleBoundsUpdateRate.Text = string.Format("Update rate in visible bounds: {0:0} updates/sec", rate*updates);

                gbStatistics.Visibility = Visibility.Visible;

                _lastStatCount = _updatesCount;
                _lastStat = DateTime.UtcNow;
            }
        }

        private void OnStartButtonClick(object sender, RoutedEventArgs e)
        {
            _updateScenarios.Add(10);
            _updateScenarios.Add(20);
            _updateScenarios.Add(50);
            _updateScenarios.Add(100);
            _updateScenarios.Add(200);
            _updateScenarios.Add(500);
            _updateScenarios.Add(1000);
            _updateScenarios.Add(5000);
            _updateScenarios.Add(10000);
            _updateScenarios.Add(20000);
            _updateScenarios.Add(50000);
            _updateScenarios.Add(100000);

            Logger.Log(string.Format("Application measures performance of Wpf Grid. It may take a little time..."));

            StartTest();
        }

        private void StartTest()
        {
            if (_updateScenarios.Count > 0)
            {
                panel.IsEnabled = false;
                int rowsCount = _updateScenarios[0];
                _updateScenarios.RemoveAt(0);

                grid.StopLoadBalancingProcess();

                grid.Headers[0].SortedColumns.Clear();


                //Populate the list of databjects
                _dataObjects.Clear();
                for (int i = 0; i < rowsCount; ++i)
                {
                    _dataObjects.Add(new TestPublisherDataObject());
                }
                
                //Populate the grid
                grid.Rows.Clear();
                foreach (TestPublisherDataObject dataObject in _dataObjects)
                {
                    grid.Rows.Add(dataObject);
                }

                if (cbEnableSort.IsChecked == true)
                {
                    grid.Headers[0][1].SortDirection = SortDirection.Ascending;
                    grid.Headers[0][3].SortDirection = SortDirection.Descending;
                }

                _measuring = false;

                grid.StartLoadBalancingProcess();
            }
            else
            {
                _measuring = false;
                panel.IsEnabled = true;

                grid.StopLoadBalancingProcess();
                panel.IsEnabled = true;

                Logger.Log(string.Format("Test completed"));
            }
        }
    }
}