﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Windows;
using Dapfor.Wpf.Controls;
using Dapfor.Wpf.Grid.Benchmark.Utils;
using Dapfor.Wpf.Utils;

namespace Dapfor.Wpf.Grid.Benchmark
{
    /// <summary>
    /// Interaction logic for RealtimeFilterControl.xaml
    /// </summary>
    public partial class RealtimeFilterControl : IActivable
    {
        readonly Random _random = new Random();
        private bool _measuring;
        private DateTime _measureStart;
        private int _updatesCount;
        private readonly List<int> _updateScenarios = new List<int>();
        readonly PerformanceCounter _cpuCounter = new PerformanceCounter();

        private DateTime _lastStat = DateTime.UtcNow;
        private int _lastStatCount;
        

        public RealtimeFilterControl()
        {
            InitializeComponent();

            //Initialize performance counters
            _cpuCounter.CategoryName = "Processor";
            _cpuCounter.CounterName = "% Processor Time";
            _cpuCounter.InstanceName = "_Total";

            grid.ProcessUpdate += OnUpdateCell;
            
            grid.Filter = new Filter(OnFilterRow);
        }

        private bool OnFilterRow(Row row)
        {
            bool filtered = false;
            foreach (Column column in grid.Headers[0].Columns)
            {
                Cell cell = row[column.Id];
                if (cell != null)
                {
                    object v = cell.Value; 
                    if (v is int)
                    {
                        int value = (int)v;
                        filtered = value < MinValue || value > MaxValue;
                        if (filtered)
                        {
                            break;
                        }    
                    }
                }
            }
            return filtered;
            
        }
   
        private void OnUpdateCell(object sender, EventArgs e)
        {
            if (grid.IsLimitFound && !_measuring)
            {
                _updatesCount = 0;
                _measuring = true;
                _measureStart = DateTime.UtcNow;
                _lastStatCount = 0;
                _lastStat = DateTime.UtcNow;
            }


            //Updates random cell in the grid;
            Row row = grid.Nodes.Count > 0 ? grid.Nodes[_random.Next() % grid.Nodes.Count] : null;
            Cell cell = row != null ? row[_random.Next() % grid.Headers[0].Count] : null;

            if (cell != null)
            {
                int value = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                if (cbNonGuiThread.IsChecked == true)
                {
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        cell.Value = value;
                    });
                }
                else
                {
                    cell.Value = value;
                }
            }

            _updatesCount++;
            if (_measuring)
            {
                //Measure refresh rate for 5 seconds
                if (DateTime.UtcNow - _measureStart > TimeSpan.FromSeconds(5))
                {
                    grid.StopLoadBalancingProcess();

                    int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;

                    long visibleCells = grid.Headers[0].Count * visibleRows;
                    long totalCells = grid.Headers[0].Count * grid.Nodes.Count;

                    double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;


                    Logger.Log(string.Format("Grid of {0} rows, size = {2:0}x{3:0}. Maximum filtering rate: {1} updates/sec, in visible bounds: {4:0} updates/sec",
                                              grid.Nodes.Count,
                                              _updatesCount / 5,
                                              grid.ActualWidth,
                                              grid.ActualHeight,
                                              rate * _updatesCount / 5));

                    //Execute the next scenario
                    StartTest();
                }
            }

            DisplayStatistics();
            
        }

        private void DisplayStatistics()
        {
            if (DateTime.UtcNow - _lastStat > TimeSpan.FromMilliseconds(500))
            {
                lbCPULoadRate.Text = string.Format("CPU load rate: {0:0} %", _cpuCounter.NextValue());
                lbGridSize.Text = string.Format("Grid size: {0:0}x{1:0}", grid.ActualWidth, grid.ActualHeight);
                double updates = 1000 * (_updatesCount - _lastStatCount) / (DateTime.UtcNow - _lastStat).TotalMilliseconds;
                lbUpdateRate.Text = string.Format("Update rate: {0:0} updates/sec", updates);

                int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                    ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                    : 0;

                long visibleCells = grid.Headers[0].Count * visibleRows;
                long totalCells = grid.Headers[0].Count * grid.Nodes.Count;

                double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;
                lbVisibleBoundsUpdateRate.Text = string.Format("Update rate in visible bounds: {0:0} updates/sec", rate * updates);

                gbStatistics.Visibility = Visibility.Visible;

                _lastStatCount = _updatesCount;
                _lastStat = DateTime.UtcNow;
            }
        }

        #region MinValueProperty

        public static readonly DependencyProperty MinValueProperty = DependencyProperty.Register("MinValue", typeof (int), typeof (RealtimeFilterControl), new PropertyMetadata(OnMinValueChanged));

        public int MinValue
        {
            get { return (int) GetValue(MinValueProperty); }
            set { SetValue(MinValueProperty, value); }
        }

        private static void OnMinValueChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = (RealtimeFilterControl) d;
            control.grid.FilterRefresh();
        }

        #endregion

        #region MaxValueProperty

        public static readonly DependencyProperty MaxValueProperty = DependencyProperty.Register("MaxValue", typeof(int), typeof(RealtimeFilterControl), new PropertyMetadata(1000, OnMaxValueChanged));

        public int MaxValue
        {
            get { return (int)GetValue(MaxValueProperty); }
            set { SetValue(MaxValueProperty, value); }
        }

        private static void OnMaxValueChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            var control = (RealtimeFilterControl)d;
            control.grid.FilterRefresh();
        }

        #endregion



        private void OnStartClick(object sender, RoutedEventArgs e)
        {
            _updateScenarios.Add(10);
            _updateScenarios.Add(20);
            _updateScenarios.Add(50);
            _updateScenarios.Add(100);
            _updateScenarios.Add(200);
            _updateScenarios.Add(500);
            _updateScenarios.Add(1000);
            _updateScenarios.Add(5000);
            _updateScenarios.Add(10000);
            _updateScenarios.Add(20000);
            _updateScenarios.Add(50000);
            _updateScenarios.Add(100000);

            Logger.Log(string.Format("Application measures performance of Wpf GridControl. It may take a little time..."));

            StartTest();
        }

        private void StartTest()
        {
            if (_updateScenarios.Count > 0)
            {
                int rowsCount = _updateScenarios[0];
                _updateScenarios.RemoveAt(0);

                grid.StopLoadBalancingProcess();

                grid.Headers[0].SortedColumns.Clear();


                //Populate the grid
                grid.Rows.Clear();
                for (int i = 0; i < rowsCount; ++i)
                {
                    TestDataObject dataObject = new TestDataObject();
                    dataObject.Value0 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value1 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value2 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value3 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value4 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value5 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value6 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value7 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value8 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));
                    dataObject.Value9 = _random.Next((int)(0.9 * MinValue), (int)(1.1 * MaxValue));

                    grid.Rows.Add(dataObject);
                }

                _measuring = false;
                panel.IsEnabled = false;
                grid.StartLoadBalancingProcess();
            }
            else
            {
                _measuring = false;
                panel.IsEnabled = true;
                grid.StopLoadBalancingProcess();

                Logger.Log(string.Format("Test completed"));
            }
        }

        public void Activate()
        {
            panel.IsEnabled = true;
        }

        public void Deactivate()
        {
            grid.StopLoadBalancingProcess();
            _updatesCount = 0;
            _measuring = true;
            _measureStart = DateTime.UtcNow;
            _lastStatCount = 0;
            _lastStat = DateTime.UtcNow;
            gbStatistics.Visibility = Visibility.Hidden;
            grid.Rows.Clear();
            panel.IsEnabled = false;
        }
    }
}
