﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using System.Windows;
using System.Windows.Media;
using Dapfor.Wpf.Controls;
using Dapfor.Wpf.Grid.Benchmark.Utils;
using Brushes = Dapfor.Wpf.Grid.Benchmark.Utils.Brushes;

namespace Dapfor.Wpf.Grid.Benchmark
{
    /// <summary>
    /// Interaction logic for RealtimeGroupingControl.xaml
    /// </summary>
    public partial class RealtimeGroupingControl : IActivable
    {
        private const int GroupCount = 20;
        readonly Random _random = new Random();
        private bool _measuring;
        private DateTime _measureStart;
        private int _updatesCount;
        private int _rowsCount;
        private readonly List<int> _updateScenarios = new List<int>();
        readonly PerformanceCounter _cpuCounter = new PerformanceCounter();
        static readonly HighPerformanceTimer HiPerfTimer = new HighPerformanceTimer();

        private DateTime _lastStat = DateTime.UtcNow;
        private int _lastStatCount;
        private bool _stopping;

        public RealtimeGroupingControl()
        {
            InitializeComponent();

            //Initialize performance counters
            _cpuCounter.CategoryName = "Processor";
            _cpuCounter.CounterName = "% Processor Time";
            _cpuCounter.InstanceName = "_Total";

            //Initialize header and columns
            grid.Headers[0].VisibleGroupPanel = true;
            grid.Headers[0][0].Grouped = true;
            grid.Headers[0][0].Visible = false;
                       

            grid.ProcessUpdate += OnUpdateCell;

           
        }

        private void OnUpdateCell(object sender, EventArgs e)
        {
            if (grid.IsLimitFound && !_measuring)
            {
                _updatesCount = 0;
                _measuring = true;
                _measureStart = DateTime.UtcNow;
                _lastStatCount = 0;
                _lastStat = DateTime.UtcNow;
            }


            //Updates random cell in the grid;
            Row row = grid.Rows.Count > 0 ? grid.Rows[_random.Next() % grid.Rows.Count] : null;
            Cell cell = row != null ? row[_random.Next() % grid.Headers[0].Count] : null;
            Brush color = _random.Next() % 2 == 0 ? Brushes.Green : Brushes.Red;


            if (cell != null)
            {
                int value = _random.Next();

                //Limit a number of groups    
                if (cell.Column.Id == "Value0")
                {
                    value = _random.Next() % GroupCount;
                }

                if (cbNonGuiThread.IsChecked == true)
                {
                    ThreadPool.QueueUserWorkItem(delegate
                    {
                        cell.Value = value;
                        cell.Blink(TimeSpan.FromMilliseconds(1000), color);
                    });
                }
                else
                {
                    cell.Value = value;
                    cell.Blink(TimeSpan.FromMilliseconds(1000), color);
                }
            }

            _updatesCount++;
            if (_measuring)
            {
                //Measure refresh rate for 5 seconds
                if (DateTime.UtcNow - _measureStart > TimeSpan.FromSeconds(5))
                {
                    grid.StopLoadBalancingProcess();

                    int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;
                    long visibleCells = grid.Headers[0].Count * visibleRows;
                    long totalCells = grid.Headers[0].Count * _rowsCount;

                    double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;


                    Logger.Log(string.Format("Grid of {0} rows, size = {2:0}x{3:0}. Maximum regrouping rate: {1} updates/sec, in visible bounds: {4:0} updates/sec",
                                              _rowsCount,
                                              _updatesCount / 5,
                                              grid.ActualWidth,
                                              grid.ActualHeight,
                                              rate * _updatesCount / 5));

                    //Execute the next scenario
                    StartTest();
                }
            }

            DisplayStatistics();
            
        }

        private void DisplayStatistics()
        {
            if (DateTime.UtcNow - _lastStat > TimeSpan.FromMilliseconds(500))
            {
                lbCPULoadRate.Text = String.Format("CPU load rate: {0:0} %", _cpuCounter.NextValue());
                lbGridSize.Text = string.Format("Grid size: {0:0}x{1:0}", grid.ActualWidth, grid.ActualHeight);
                double updates = 1000 * (_updatesCount - _lastStatCount) / (DateTime.UtcNow - _lastStat).TotalMilliseconds;
                lbUpdateRate.Text = string.Format("Update rate: {0:0} updates/sec", updates);

                int visibleRows = grid.Rows.BottomVisibleRow != null && grid.Rows.TopVisibleRow != null
                        ? (grid.Rows.BottomVisibleRow.VisibleIndex - grid.Rows.TopVisibleRow.VisibleIndex)
                        : 0;

                long visibleCells = grid.Headers[0].Count * visibleRows;
                long totalCells = grid.Headers[0].Count * grid.Rows.Count;

                double rate = totalCells != 0 ? (double)visibleCells / totalCells : 0;
                lbVisibleBoundsUpdateRate.Text = string.Format("Update rate in visible bounds: {0:0} updates/sec", rate * updates);

                gbStatistics.Visibility = Visibility.Visible;

                _lastStatCount = _updatesCount;
                _lastStat = DateTime.UtcNow;
            }

        }

        private void OnStartClick(object sender, RoutedEventArgs e)
        {
            panel.IsEnabled = false;
            _stopping = false;


            const int tranche = 10000;

            Logger.Log(string.Format("Performance measurement when Wpf Grid groups rows."));
            Helper.DoEvents();

            //10 iterations
            for (int iteration = 0; iteration < 10 && !_stopping; ++iteration)
            {
                grid.Headers[0].SortedColumns.Clear();
                grid.Rows.Clear();
                grid.Headers[0].GroupedColumns.Clear();


                //Add rows to the grid non-sorted and non-grouped grid
                for (int i = 0; i < (iteration + 1) * tranche && !_stopping; ++i)
                {
                    //Limit a number of groups, otherwise the grid will have a group per row
                    TestDataObject dataObject = new TestDataObject();
                    dataObject.Value0 = _random.Next() % GroupCount;
                    grid.Rows.Add(dataObject);
                }

                HiPerfTimer.Start();

                //Group by the first column
                grid.Headers[0]["Value0"].Grouped = true;
                grid.Headers[0]["Value0"].Visible = false;

                HiPerfTimer.Stop();

                int rate = HiPerfTimer.Duration > 0 ? (int)(1000 * ((iteration + 1) * tranche) / HiPerfTimer.Duration) : 0;
                Logger.Log(string.Format("Grid grouped {0} rows in {1:0} msec. Rate: {2} rows/sec", (iteration + 1) * tranche, HiPerfTimer.Duration, rate));
                //grid.Rows.ExpandAll();
                Helper.DoEvents();
            }


            panel.IsEnabled = true;
            Logger.Log(string.Format("Test completed"));
           
        }

        private void OnStartRealTimeClick(object sender, RoutedEventArgs e)
        {
            _updateScenarios.Add(10);
            _updateScenarios.Add(20);
            _updateScenarios.Add(50);
            _updateScenarios.Add(100);
            _updateScenarios.Add(200);
            _updateScenarios.Add(500);
            _updateScenarios.Add(1000);
            _updateScenarios.Add(5000);
            _updateScenarios.Add(10000);
            _updateScenarios.Add(20000);
            _updateScenarios.Add(50000);
            _updateScenarios.Add(100000);

            Logger.Log(string.Format("Application measures performance of Wpf Grid while real-time regrouping. It may take a little time..."));

            _stopping = false;
            StartTest();
        }

        public void Activate()
        {
            panel.IsEnabled = true;
        }

        public void Deactivate()
        {
            grid.StopLoadBalancingProcess();
            _updatesCount = 0;
            _measuring = true;
            _measureStart = DateTime.UtcNow;
            _lastStatCount = 0;
            _lastStat = DateTime.UtcNow;
            gbStatistics.Visibility = Visibility.Hidden;
            grid.Rows.Clear();
            panel.IsEnabled = false;
            _stopping = true;
        }

        void StartTest()
        {
            if (_updateScenarios.Count > 0 && !_stopping)
            {
                _rowsCount = _updateScenarios[0];
                _updateScenarios.RemoveAt(0);

                grid.StopLoadBalancingProcess();

                grid.Headers[0].SortedColumns.Clear();


                //Populate the grid
                grid.Rows.Clear();
                for (int i = 0; i < _rowsCount && !_stopping; ++i)
                {
                    //Limit a number of groups, otherwise the grid will have a group per row
                    TestDataObject dataObject = new TestDataObject();
                    //dataObject.Value0 = random.Next() % groupCount;
                    int groups = Math.Min(_rowsCount, GroupCount);

                    dataObject.Value0 = i / (_rowsCount / groups);

                    grid.Rows.Add(dataObject);
                }
                //grid.Rows.ExpandAll();

                _measuring = false;
                panel.IsEnabled = false;
                grid.StartLoadBalancingProcess();
            }
            else
            {
                _measuring = false;
                panel.IsEnabled = true;
                grid.StopLoadBalancingProcess();

                Logger.Log(string.Format("Test completed"));
            }
        }
    }
}